from __future__ import print_function

from django import forms
from django.contrib.admin.widgets import ManyToManyRawIdWidget, ForeignKeyRawIdWidget
from django.core.urlresolvers import reverse, NoReverseMatch
from django.forms.widgets import Select, TextInput
try:
    from django.forms.widgets import flatatt
except ImportError:
    # Changed in Django>=1.11.
    from django.forms.utils import flatatt
from django.template import Context, Template
from django.template.context import Context
from django.utils.encoding import force_text, smart_text
from django.utils.html import escape
from django.utils.safestring import mark_safe

try:
    from django.utils.encoding import StrAndUnicode # pylint: disable=unused-import
except ImportError:
    from django.utils.encoding import python_2_unicode_compatible

    @python_2_unicode_compatible
    class StrAndUnicode:
        def __str__(self):
            return self.code

from admin_steroids import utils

class LinkedSelect(Select):
    def render(self, name, value, attrs=None, *args, **kwargs):
        output = super(LinkedSelect, self).render(name, value, attrs=attrs, *args, **kwargs)
        model = self.choices.field.queryset.model
        to_field_name = self.choices.field.to_field_name or 'id'
        try:
            kwargs = {to_field_name:value}
            obj = model.objects.get(**kwargs)
            view_url = utils.get_admin_change_url(obj)
            output += mark_safe('&nbsp;<a href="%s" target="_blank">view</a>&nbsp;' % (view_url,))
        except model.DoesNotExist:
            pass
        return output

class ForeignKeyTextInput(TextInput):
    """
    Implements the same markup as VerboseForeignKeyRawIdWidget but does not
    require an explicit model relationship.
    """

    def __init__(self, model_class, value, *args, **kwargs):
        super(ForeignKeyTextInput, self).__init__(*args, **kwargs)
        self._model_class = model_class
        try:
            value = int(value)
        except ValueError:
            value = 0
        except TypeError:
            value = 0
        self._raw_value = value
        q = model_class.objects.filter(id=value)
        self._instance = None
        if q.count():
            self._instance = q[0]

    def render(self, name, value, attrs=None, renderer=None):
        if value is None:
            value = ''
        final_attrs = self.build_attrs(attrs)
        if value != '':
            # Only add the 'value' attribute if a value is non-empty.
            final_attrs['value'] = force_text(self._format_value(value))
        final_attrs['size'] = 10
        final_attrs['type'] = 'text'
        final_attrs['name'] = name
        t = Template(u"""
{% load staticfiles %}
<input{{ attrs|safe }} />
{% if instance %}
    <a href="{{ changelist_url|safe }}?t=id" class="related-lookup" id="lookup_{{ id|safe }}" onclick="return showRelatedObjectLookupPopup(this);">
        <img src="{% static 'admin/img/selector-search.gif' %}" width="16" height="16" alt="Lookup" />
    </a>
    <strong><a href="{{ url|safe }}" target="_blank">{{ instance|safe }}</a></strong>
{% endif %}
        """)
        c = Context(dict(
            id=final_attrs['id'],
            attrs=flatatt(final_attrs),
            raw_value=self._raw_value,
            url=utils.get_admin_change_url(self._instance),
            changelist_url=utils.get_admin_changelist_url(self._model_class),
            instance=self._instance,
        ))
        return mark_safe(t.render(c))

#http://djangosnippets.org/snippets/2217/

class VerboseForeignKeyRawIdWidget(ForeignKeyRawIdWidget):

    def __init__(self, *args, **kwargs):
        raw_id_fields_new_tab = True
        if 'raw_id_fields_new_tab' in kwargs:
            raw_id_fields_new_tab = kwargs['raw_id_fields_new_tab']
            del kwargs['raw_id_fields_new_tab']
        super(VerboseForeignKeyRawIdWidget, self).__init__(*args, **kwargs)
        self.raw_id_fields_new_tab = raw_id_fields_new_tab

    @property
    def target(self):
        if self.raw_id_fields_new_tab:
            return '_blank'
        return '_self'

    def label_for_value(self, value):
        key = self.rel.get_related_field().name
        try:
            obj = self.rel.to._default_manager.using(self.db).get(**{key: value})
            change_url = reverse("admin:%s_%s_change" % (obj._meta.app_label, obj._meta.object_name.lower()), args=(obj.pk,))
            return '&nbsp;<strong><a href="%s" target="%s">%s</a></strong>' % (change_url, self.target, escape(obj))
        except NoReverseMatch:
            return '&nbsp;<strong>%s</strong>' % (escape(obj),)
        except (ValueError, self.rel.to.DoesNotExist):
            return ''

class VerboseManyToManyRawIdWidget(ManyToManyRawIdWidget):

    def __init__(self, *args, **kwargs):
        raw_id_fields_new_tab = True
        if 'raw_id_fields_new_tab' in kwargs:
            raw_id_fields_new_tab = kwargs['raw_id_fields_new_tab']
            del kwargs['raw_id_fields_new_tab']
        super(VerboseManyToManyRawIdWidget, self).__init__(*args, **kwargs)
        self.raw_id_fields_new_tab = raw_id_fields_new_tab

    @property
    def target(self):
        if self.raw_id_fields_new_tab:
            return '_blank'
        return '_self'

    def label_for_value(self, value):
        values = value.split(',')
        str_values = []
        key = self.rel.get_related_field().name
        for v in values:
            obj = self.rel.to._default_manager.using(self.db).get(**{key: v})
            x = smart_text(obj)
            try:
                change_url = reverse("admin:%s_%s_change" % (obj._meta.app_label, obj._meta.object_name.lower()), args=(obj.pk,))
                str_values += ['<strong><a href="%s" target="%s">%s</a></strong>' % (change_url, self.target, escape(x))]
            except NoReverseMatch:
                str_values += ['<strong>%s</strong>' % (escape(x),)]
            except self.rel.to.DoesNotExist:
                str_values += [u'???']
        return u', '.join(str_values)

class PlainTextWidget(forms.Widget):
    """
    Renders the value as plain text.
    """
    def render(self, _name, value, attrs=None, renderer=None):
        value = value or ''
        return mark_safe('<div style="padding-top:3px;">'+value+'</div>')

class PreTextWidget(forms.Widget):
    """
    Renders the value as plain text formatted with the "pre" style.
    """
    def render(self, _name, value, attrs=None, renderer=None):
        value = value or ''
        return mark_safe('<div style="padding-top:3px; white-space:pre;">'+value+'</div>')

class NBSPTextWidget(forms.Widget):
    """
    Renders the value as plain text with all spaces replaced by "&nbsp;".
    """
    def render(self, _name, value, attrs=None, renderer=None):
        value = value or ''
        value = value.replace(' ', '&nbsp;').replace('\n', '<br/>')
        return mark_safe('<div style="padding-top:3px;">'+value+'</div>')

class BRTextWidget(forms.Widget):
    """
    Renders the value as plain text with all newlines replaced by "<br/>".
    """
    def render(self, _name, value, attrs=None, renderer=None):
        value = value or ''
        value = value.replace('\n', '<br/>')
        _attrs = self.attrs.copy()
        _attrs.update(attrs or {})
        style = _attrs.get('style', '')
        return mark_safe('<div style="'+style+'">'+value+'</div>')

class ReadOnlyText(forms.TextInput):
    """
    Renders the value as plain text.
    """
    def render(self, name, value, attrs=None, renderer=None):
        if value is None:
            value = ''
        return value



from __future__ import print_function

import uuid

from django.contrib.admin import FieldListFilter, SimpleListFilter, ListFilter
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.core.urlresolvers import reverse
from django.core.cache import cache
from django.db import models
from django.db.models import Q
from django.utils.translation import ugettext as _
from django.utils.encoding import smart_text
from django.contrib.admin.options import IncorrectLookupParameters
from django.contrib.auth.models import User

def get_empty_value_display(cl):
    if hasattr(cl.model_admin, 'get_empty_value_display'):
        return cl.model_admin.get_empty_value_display()
    # Django < 1.9
    from django.contrib.admin.views.main import EMPTY_CHANGELIST_VALUE # pylint: disable=no-name-in-module
    return EMPTY_CHANGELIST_VALUE

class NullListFilter(FieldListFilter):

    def __init__(self, field, request, params, model, model_admin, field_path):
        self.lookup_kwarg = '%s__isnull' % field_path
        self.lookup_val = None
        try:
            self.lookup_val = request.GET.get(self.lookup_kwarg, None)
            if self.lookup_val is not None:
                if self.lookup_val in (True, 'True', 1, '1'): # pylint: disable=R0102
                    self.lookup_val = True
                else:
                    self.lookup_val = False
        except Exception as e:
            pass
        super(NullListFilter, self).__init__(field,
            request, params, model, model_admin, field_path)

    def expected_parameters(self):
        return [self.lookup_kwarg,]

    def choices(self, cl):
        for lookup, title in (
                (None, _('All')),
                (False, _('Has value')),
                (True, _('Omitted'))):
            d = {
                'selected': self.lookup_val == lookup,
                'query_string': cl.get_query_string({
                    self.lookup_kwarg: lookup,
                }, [self.lookup_kwarg]),
                'display': title,
            }
            yield d

class NullBlankListFilter(FieldListFilter):
    """
    Like NullListFilter, but treats None and '' values synonymously.
    """

    def __init__(self, field, request, params, model, model_admin, field_path):
        self.field_path = field_path
        self.lookup_kwarg = '%s_isnullblank' % field_path

        self.lookup_val = None
        try:
            self.lookup_val = request.GET.get(self.lookup_kwarg, None)
            if self.lookup_val is not None:
                if self.lookup_val in (True, 'True', 1, '1'): # pylint: disable=R0102
                    self.lookup_val = True
                else:
                    self.lookup_val = False
        except Exception as e:
            pass

        super(NullBlankListFilter, self).__init__(field,
            request, params, model, model_admin, field_path)

    def expected_parameters(self):
        return [self.lookup_kwarg]

    def queryset(self, request, queryset):
        try:
            if self.lookup_val is True:
                queryset = queryset.filter(
                    Q(**{self.field_path+'__isnull': True})|\
                    Q(**{self.field_path: ''}))
            elif self.lookup_val is False:
                queryset = queryset.exclude(
                    Q(**{self.field_path+'__isnull': True})|\
                    Q(**{self.field_path: ''}))
            return queryset
        except ValidationError as e:
            raise IncorrectLookupParameters(e)

    def choices(self, cl):
        for lookup, title in (
                (None, _('All')),
                (False, _('Has value')),
                (True, _('Omitted'))):
            d = {
                'selected': self.lookup_val == lookup,
                'query_string': cl.get_query_string({
                    self.lookup_kwarg: lookup,
                }, [self.lookup_kwarg]),
                'display': title,
            }
            yield d

class NotInListFilter(FieldListFilter):
    """
    Allows the use of exclude(field=value) via the URL.
    The inverse of Django's default "__in=" URL syntax.
    """

    def __init__(self, field, request, params, model, model_admin, field_path):
        self.field_path = field_path
        self.lookup_kwarg = '%s__notin' % field_path
        self.lookup_vals = None
        try:
            self.lookup_vals = request.GET.get(self.lookup_kwarg, None)
            if self.lookup_vals is not None:
                self.lookup_vals = self.lookup_vals.split(',')
        except Exception as e:
            pass

        self.lookup_choices = field.get_choices(include_blank=False)
        super(NotInListFilter, self).__init__(field,
            request, params, model, model_admin, field_path)

        self.title = getattr(field, 'verbose_name', field_path) + ' is not'

    def expected_parameters(self):
        return [self.lookup_kwarg]

    def queryset(self, request, queryset):
        try:
            # Convert the __notin to a Django ORM .exclude(...)
            if self.lookup_kwarg in self.used_parameters and self.lookup_vals:
                queryset = queryset.exclude(**{self.field_path+'__in': self.lookup_vals})
            return queryset
        except ValidationError as e:
            raise IncorrectLookupParameters(e)

    def choices(self, cl):
        yield {
            'selected': self.lookup_vals is None,# and not self.lookup_val_isnull,
            'query_string': cl.get_query_string({},
                [self.lookup_kwarg]),
            'display': _('None'),
        }
        for pk_val, val in self.lookup_choices:
            yield {
                'selected': (smart_text(pk_val) in self.lookup_vals) if self.lookup_vals else False,
                'query_string': cl.get_query_string({
                    self.lookup_kwarg: pk_val,
                }, []),
                'display': val,
            }

class CachedFieldFilter(FieldListFilter):
    """
    Caches the choices query from the model, ignoring any other filtering
    on the model.
    """

    cache_seconds = 3600 # 1-hour

    def __init__(self, field, request, params, model, model_admin, field_path):
        self.field_name = field.name
        self.lookup_kwarg = '%s__exact' % field_path
        self.lookup_kwarg2 = '%s__isnull' % field_path
        self.lookup_val = request.GET.get(self.lookup_kwarg, None)
        self.lookup_val2 = request.GET.get(self.lookup_kwarg2, None)

        super(CachedFieldFilter, self).__init__(
            field, request, params, model, model_admin, field_path)

        self.model = model

    def expected_parameters(self):
        return [self.lookup_kwarg, self.lookup_kwarg2]

    def choices(self, cl):
        # Query cached choices.
        # Note, this purposefully gets a distinct set from the global
        # set of values, so that when we cache it, it's valid for all
        # admin queries. Yes, it may include some values that will return
        # no results on some pages, but that's an acceptable trade-off for
        # being able to shave off a lot of query time.
        cache_key = 'cff_%s_%s' \
            % (str(self.model).split("'")[1], self.field_name)
        values = cache.get(cache_key)
        if values is None:
            values = self.model.objects.all()\
                .values_list(self.field_name, flat=True)\
                .distinct().order_by(self.field_name)
            cache.set(cache_key, values, self.cache_seconds)

        yield {
            'selected': self.lookup_val is None and self.lookup_val2 is None,
            'query_string': cl.get_query_string({
                    self.lookup_kwarg2: '',
                }, [self.lookup_kwarg]),
            'display': _('All'),
        }

        for value in values:
            if value is None:
                yield {
                    'selected': self.lookup_val2,
                    'query_string': cl.get_query_string({
                            self.lookup_kwarg: value,
                        }, [self.lookup_kwarg2]),
                    'display': value,
                }
            else:
                yield {
                    'selected': self.lookup_val == value,
                    'query_string': cl.get_query_string({
                            self.lookup_kwarg: value,
                        }, [self.lookup_kwarg2]),
                    'display': value,
                }

class AjaxFieldFilter(FieldListFilter):
    """
    Allows searching for one or more field values via ajax
    and searching for them via __exact or __in.
    """

    template = 'admin_steroids/ajax_filter.html'

    #TODO:specify one-only or multiple
    def __init__(self, field, request, params, model, model_admin, field_path):
        self.field_name = field.name
        self.field = field

        self.lookup_kwarg = '%s__in' % field_path
        self.lookup_val = [
            _ for _ in request.GET.get(self.lookup_kwarg, '').split(',')
            if _.strip()
        ]

        super(AjaxFieldFilter, self).__init__(
            field, request, params, model, model_admin, field_path)

        self.model = model

        self.uuid = '_'+str(uuid.uuid4())

#        _d = request.GET.copy()
#        del _d[self.lookup_kwarg]
#        self.base_url = request.path + '?' + _d.urlencode
#        if _d:
#            self.base_url += '&'
#        self.base_url += self.lookup_kwarg + '='

        self.ajax_url = reverse(
            'model_field_search',
            args=(model._meta.app_label, model.__name__.lower(), self.field_name))

#    def __call__(self, *args, **kwargs):
#        return

    def expected_parameters(self):
        return [self.lookup_kwarg]

    @property
    def values(self):
        return self.lookup_val

    def choices(self, cl):
        # Note, all these choices are for *deselecting* the value.
        # Additions will be handled dynamically via AJAX.
        yield {
            'selected': not self.lookup_val,
            'query_string': cl.get_query_string(
                new_params={},
                remove=[self.lookup_kwarg],
            ),
            'display': _('All'),
            'remove_icon': False,
            'alt': 'Remove All',
        }
        for value in self.values:
            lst_without = list(self.lookup_val)
            lst_without.remove(value)

            if lst_without:
                url = cl.get_query_string(
                    new_params={self.lookup_kwarg: ','.join(lst_without)},
                    remove=[self.lookup_kwarg],
                )
            else:
                url = cl.get_query_string(
                    new_params={},
                    remove=[self.lookup_kwarg],
                )

            # Lookup the "pretty" display value for IDs of related models.
            if isinstance(self.field,
                (
                    models.ForeignKey,
                    models.ManyToManyField,
                    models.OneToOneField,
                )):
                rel_model = self.field.rel.to
                #TODO:handle non-numeric IDs?
                # AvB edit here; some FK models do not have an (integer) or .id attribute!
                #value = str(rel_model.objects.get(id=int(value)))
                if value.isdigit():
                    value = int(value)
                value = str(rel_model.objects.get(pk=value))

            yield {
                'selected': True,
                'query_string': url,
                'display': value,
                'remove_icon': True,
                'alt': 'Remove',
            }

class LogEntryAdminUserFilter(SimpleListFilter):
    title = _('user')
    parameter_name = 'user'

    def lookups(self, request, model_admin):
        qs = User.objects.filter(Q(is_staff=True) | Q(is_superuser=True))
        qs = qs.order_by('username')
        return [(user.pk, _(str(user))) for user in qs]

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(user__pk=self.value())

        return queryset

#http://stackoverflow.com/a/20588975/247542
class SingleTextInputFilter(ListFilter):
    """
    renders filter form with text input and submit button
    """
    size = 40
    parameter_name = None
    template = 'admin_steroids/textinput_filter.html'
    style = ''

    def __init__(self, request, params, model, model_admin):
        super(SingleTextInputFilter, self).__init__(
            request, params, model, model_admin)
        if self.parameter_name is None:
            raise ImproperlyConfigured(
                "The list filter '%s' does not specify "
                "a 'parameter_name'." % self.__class__.__name__)

        if self.parameter_name in params:
            value = params.pop(self.parameter_name)
            self.used_parameters[self.parameter_name] = value

    def value(self):
        """
        Returns the value (in string format) provided in the request's
        query string for this filter, if any. If the value wasn't provided then
        returns None.
        """
        return self.used_parameters.get(self.parameter_name, None)

    def has_output(self):
        return True

    def expected_parameters(self):
        """
        Returns the list of parameter names that are expected from the
        request's query string and that will be used by this filter.
        """
        return [self.parameter_name]

    def queryset(self, request, queryset):
        raise NotImplementedError

    def choices(self, cl):
        all_choice = {
            'selected': self.value() is None,
            'query_string': cl.get_query_string({}, [self.parameter_name]),
            'display': _('All'),
        }
        v = self.value()
        return ({
            'get_query': cl.params,
            'current_value': '' if v is None else v,
            'all_choice': all_choice,
            'parameter_name': self.parameter_name,
            'query_string': all_choice['query_string'],
            'title': self.title,
            'size': self.size,
            'style': self.style,
        }, )



"""
Quick test:

    export TESTNAME=.test_delete_duplicate_record; tox -e py27-django15

"""
from __future__ import print_function

import time
import socket
import warnings
import csv

from django.core import mail
from django.test import TestCase
try:
    from django.test import override_settings
except ImportError:
    from override_settings import override_settings

# pylint: disable=C0412
from admin_steroids import utils

warnings.simplefilter('error', RuntimeWarning)

socket.gethostname = lambda: 'localhost'

# Test settings that we don't want to use for all tests.
EMAIL_HOST = 'localhost'
EMAIL_PORT = '1025'
DEV_EMAIL_REDIRECT_TO = 'test@domain.com'
SMTPD_LOG = '/tmp/smtpd.log'

class Tests(TestCase):

#     fixtures = []

    def setUp(self):
        pass

    def test_obj_to_hash(self):
        s = utils.obj_to_hash({123:'abc'})
        self.assertEqual(len(s), 128)

    def test_FormatWithCommas(self):
        s = utils.FormatWithCommas('%.4f', 1234567.5678)
        self.assertEqual(s, '1,234,567.5678')

    @override_settings(EMAIL_BACKEND='admin_steroids.email.DevelopmentEmailBackend')
    @override_settings(EMAIL_HOST=EMAIL_HOST)
    @override_settings(EMAIL_PORT=EMAIL_PORT)
    @override_settings(DEV_EMAIL_REDIRECT_TO=DEV_EMAIL_REDIRECT_TO)
    @override_settings(DEV_EMAIL_APPEND_HOSTNAME=True)
    @override_settings(DEV_EMAIL_ALLOW_ANY_ON_DOMAIN=False)
    def _test_DevelopmentEmailBackend(self):

        # Kill all previous SMTP servers.
#         os.system("ps aux|grep -i smtp|grep -v grep|awk '{print $2}' | xargs -i kill {}")

        # Launch debuggin SMTP server to catch emails.
#         print('a')
#         command = 'python -m smtpd -n -c DebuggingServer %s:%s > %s' \
#             % (EMAIL_HOST, EMAIL_PORT, SMTPD_LOG)
#         print('command:', command)
#         process = subprocess.Popen(command, shell=True)
#         time.sleep(1)
#         print('b')

        output = []

        try:
            #wait_until_response(num=1)

            # Send email.
            self.assertEqual(len(mail.outbox), 0)
            mail.send_mail(
                subject='Subject here',
                message='Here is the message.',
                from_email='from@example.com',
                recipient_list=['to@example.com'],
                fail_silently=False,
            )
            time.sleep(3)

            # Confirm the email didn't go into Django's fake email backend.
            self.assertEqual(len(mail.outbox), 0)

#             last = None
#             t0 = time.time()
#             timeout = 10
#             while 1:
#                 print('checking smtp')
#                 chunk = open(SMTPD_LOG, 'r').read().strip()
#                 if last is not None and last == chunk:
#                     break
#                 last = chunk
#                 time.sleep(1)
#                 if time.time() - t0 > timeout:
#                     break

            # Poll process for new output until finished
            output = '\n'.join(output)
            print('output:', output)

            signature = '(Sent from http://%s)' % EMAIL_HOST
            self.assertTrue(signature in output)

        finally:

            print('Killing SMTP...')
#             if process:
#                 process.kill()
            print('Killed SMTP.')

    def test_csv_encoding(self):
        s = u'\ufffd'

        with self.assertRaises(UnicodeEncodeError):
            print(s.encode('ascii'))

        with self.assertRaises((UnicodeEncodeError, TypeError)):
            with open('/tmp/test.csv', 'wb') as fout:
                writer = csv.DictWriter(fout, fieldnames=['name'])
                writer.writerow({'name': s})

        with open('/tmp/test.csv', 'w') as fout:
            writer = csv.DictWriter(fout, fieldnames=['name'])
            data = utils.encode_csv_data({'name': s})
            print('data:', data)
            writer.writerow(data)

    def test_delete_duplicate_record(self):
        from admin_steroids.tests.models import Person, Contact
        from django.core.management import call_command

        print('Confirming records with no clear conflicts can be merged...')
        Person.objects.all().delete()
        p3 = Person.objects.create(name='Sue')
        Contact.objects.create(person=p3, email='sue@sue.com')
        p4 = Person.objects.create(name='Susan')
        self.assertEqual(Person.objects.all().count(), 2)
        self.assertEqual(Contact.objects.all().count(), 1)
        call_command('delete_duplicate_record', 'tests.person', p4.id, p3.id)
        self.assertEqual(Person.objects.all().count(), 1)
        self.assertEqual(Contact.objects.all().count(), 1)

        print('Confirming records with conflicting dependencies cannot be merged...')
        Person.objects.all().delete()
        self.assertEqual(Person.objects.all().count(), 0)
        self.assertEqual(Contact.objects.all().count(), 0)
        p1 = Person.objects.create(name='Bob')
        Contact.objects.create(person=p1, email='bob@bob.com')
        p2 = Person.objects.create(name='Bobby')
        Contact.objects.create(person=p2, email='bob@bob.com')
        call_command('delete_duplicate_record', 'tests.person', p1.id, p2.id)
        self.assertEqual(Person.objects.all().count(), 2)
        self.assertEqual(Contact.objects.all().count(), 2)

    def test_widgets(self):
        import django
        print('django.version:', django.VERSION)
        from admin_steroids import widgets

    def test_currency(self):
        from admin_steroids.fields import Currency
        value = Currency('$500,000.00')
        self.assertEqual(value, 500000)
        self.assertEqual(value, 500000.0)
        self.assertEqual(value.format(), '500,000.00')



from __future__ import print_function

import sys
#import traceback
from collections import defaultdict
from optparse import make_option

from django import get_version, VERSION
from django.core.management.base import BaseCommand
from django.contrib.contenttypes.models import ContentType

try:
    from django.test import override_settings
except ImportError:
    from override_settings import override_settings

from admin_steroids.queryset import atomic

def get_options(parser=None):
    make_opt = make_option
    if parser:
        make_opt = parser.add_argument
    return [
        make_opt('--dryrun', action='store_true', default=False),
        make_opt('--only-show-classes', action='store_true', default=False),
        make_opt('--do-update', action='store_true', default=False,
            help='If given, does a SQL update instead of calling the model\'s save() method. '
                'This is only recommended for use when there are circular FK references coupled '
                'with validation logic preventing incremental saves.'),
    ]

def get_all_related_objects(obj):
    try:
        links = obj._meta.get_all_related_objects()
    except AttributeError:
        # get_all_related_objects() was removed in Django >= 1.9
        # https://docs.djangoproject.com/es/1.9/ref/models/meta/
        links = [
            f for f in obj._meta.get_fields()
            if (f.one_to_many or f.one_to_one) and f.auto_created and not f.concrete
        ]
    return links

class Command(BaseCommand):
    help = 'Replaces one record with another, making sure to update all foreign key references.'
    args = 'app.model old_id new_id'
    option_list = getattr(BaseCommand, 'option_list', ()) + tuple(get_options())

    def create_parser(self, prog_name, subcommand):
        """
        For ``Django>=1.10``
        Create and return the ``ArgumentParser`` which extends ``BaseCommand`` parser with
        chroniker extra args and will be used to parse the arguments to this command.
        """
        from distutils.version import StrictVersion # pylint: disable=E0611,import-error
        parser = super(Command, self).create_parser(prog_name, subcommand)
        version_threshold = StrictVersion('1.10')
        current_version = StrictVersion(get_version(VERSION))
        if current_version >= version_threshold:
            parser.add_argument('name')
            parser.add_argument('old_id')
            parser.add_argument('new_id')
            get_options(parser)
            self.add_arguments(parser)
        return parser

    @atomic
    @override_settings(DEBUG=False)
    def handle(self, name, old_id, new_id, **options):

        dryrun = options['dryrun']
        only_show_classes = options['only_show_classes']
        do_update = options['do_update']

        app_label, model_name = name.split('.')
        ct = ContentType.objects.get(app_label=app_label, model=model_name)
        model_cls = ct.model_class()

        old_obj = model_cls.objects.get(pk=int(old_id))
        new_obj = model_cls.objects.get(pk=int(new_id))

        print('Attempting to replace %s with %s...' % (old_obj, new_obj))

        deleted_objects = set()
        # [(new_obj, old_obj, referring_obj, referring_field, exception)]
        deletion_exceptions = set()
        deletion_failures = 0
        safe_to_delete = True
        referring_classes = defaultdict(int)
        links = get_all_related_objects(old_obj)
        print('%i links found.' % len(links))
        for link in links:
            if not link.model._meta.managed:
                continue
            try:
                referring_objects = getattr(old_obj, link.get_accessor_name()).all()
                total = referring_objects.count()
                referring_objects_iters = referring_objects.iterator()
            except AttributeError:
                referring_objects = getattr(old_obj, link.get_accessor_name())
                if referring_objects:
                    total = 1
                    referring_objects_iters = [referring_objects]
                else:
                    total = 0
                    referring_objects_iters = []
            i = 0
            print('%i referring objects found on link %s.' % (total, link.get_accessor_name()))
            for referring_object in referring_objects_iters:
                i += 1

                if only_show_classes:
                    referring_classes[link.model.__name__] += 1
                    continue

                print('Changing %s(id=%i).%s = "%s"(%i) -> "%s"(%i). (%i of %i)' % (
                    link.model.__name__,
                    referring_object.id,
                    link.field.name,
                    getattr(referring_object, link.field.name),
                    getattr(referring_object, link.field.name).id,
                    new_obj,
                    new_obj.id,
                    i,
                    total,
                ))
                deleted_objects.add(
                    (type(old_obj).__name__, old_obj.id, new_obj)
                )
                try:
                    if do_update:
                        # Bypass save() logic and directly update the FK field.
                        type(referring_object).objects\
                            .filter(pk=referring_object.pk)\
                            .update(**{link.field.name: new_obj})
                    else:
                        # Set field and then save through the ORM.
                        setattr(referring_object, link.field.name, new_obj)
                        referring_object.save()
                except Exception as e:
                    print(e, file=sys.stderr)
                    safe_to_delete = False
                    deletion_exceptions.add((
                        new_obj,
                        old_obj,
                        referring_object,
                        link.field.name,
                        e,
                    ))

        if only_show_classes:
            print('Classes referring to %s:' % old_obj)
            for _mdl, _cnt in referring_classes.iteritems():
                print(_cnt, _mdl)
        else:
            # Now all FK links should be gone so we can safely delete the duplicate.
            if safe_to_delete:
                old_obj.delete()
            else:
                deletion_failures += 1

        if dryrun:
            print('%i objects pending deletion.' % len(deleted_objects))
            deleted_by_type = {}
            for cls_name, deleted_obj_id, real_obj in deleted_objects:
                deleted_by_type.setdefault(cls_name, [])
                deleted_by_type[cls_name].append((deleted_obj_id, real_obj.id))
            for cls_name, id_lst in deleted_by_type.iteritems():
                print(cls_name, ', '.join(
                    '%s -> %s' % (old_id, new_id) for old_id, new_id in id_lst
                ))
        else:
            print('%i objects deleted.' % len(deleted_objects))

        if deletion_failures:
            print('!'*80)
            print('%i deletion failures!' % deletion_failures)
            for del_exc in deletion_exceptions:
                new_obj, dup_obj, other_instance, other_field_name, exc = del_exc
                print('Unable to change %s(id=%i).%s from %s(%s) to %s(%s): %s' % (
                    type(other_instance).__name__,
                    other_instance.id,
                    other_field_name,
                    dup_obj, dup_obj.id,
                    new_obj, new_obj.id,
                    exc,
                ))



from __future__ import print_function

import csv
from inspect import isclass

from django.contrib import admin
from django.contrib.admin.sites import site
from django.forms.models import ModelForm
from django.http import HttpResponse
from django.template.defaultfilters import slugify
from django.utils.safestring import mark_safe

import six

from .utils import get_admin_change_url
from . import widgets as w
from . import utils
from . import filters

class BaseModelAdmin(admin.ModelAdmin):

#    # Cleanup the breadcrumbs on the change page.
#    def change_view(self, request, object_id, form_url='', extra_context=None):
#        extra_context = extra_context or {}
#        extra_context['app_label'] = self.model._meta.app_label.title()
#        return super(BaseModelAdmin, self).change_view(request, object_id, form_url, extra_context)

    # Cleanup the breadcrumbs on the changelist page.
    def changelist_view(self, request, extra_context=None):
        extra_context = extra_context or {}
        extra_context['app_label'] = self.model._meta.app_label.title()
        return super(BaseModelAdmin, self).changelist_view(request, extra_context)

    # Cleanup the breadcrumbs on the delete page.
    def delete_view(self, request, object_id, extra_context=None):
        extra_context = extra_context or {}
        extra_context['app_label'] = self.model._meta.app_label.title()
        return super(BaseModelAdmin, self).delete_view(request, object_id, extra_context)

# Based on http://djangosnippets.org/snippets/2217/.
class BetterRawIdFieldsModelAdmin(BaseModelAdmin):
    """
    Displays all raw id fields in a modeladmin as a link going to that record's
    associated admin change page.
    """

    raw_id_fields_new_tab = True

    def formfield_for_dbfield(self, db_field, **kwargs):
        if db_field.name in self.raw_id_fields:
            kwargs.pop("request", None)
            typ = db_field.rel.__class__.__name__
            if typ == "ManyToOneRel" or typ == "OneToOneRel":
                kwargs['widget'] = w.VerboseForeignKeyRawIdWidget(db_field.rel, site, raw_id_fields_new_tab=self.raw_id_fields_new_tab)
            elif typ == "ManyToManyRel":
                kwargs['widget'] = w.VerboseManyToManyRawIdWidget(db_field.rel, site, raw_id_fields_new_tab=self.raw_id_fields_new_tab)
            return db_field.formfield(**kwargs)
        return super(BetterRawIdFieldsModelAdmin, self).formfield_for_dbfield(db_field, **kwargs)

ImproveRawIdFieldsForm = BetterRawIdFieldsModelAdmin

class BetterRawIdFieldsTabularInline(admin.TabularInline):
    """
    Displays all raw id fields in a tabular inline as a link going to that
    record's associated admin change page.
    """

    raw_id_fields_new_tab = True

    def formfield_for_dbfield(self, db_field, **kwargs):
        if db_field.name in self.raw_id_fields:
            kwargs.pop("request", None)
            typ = db_field.rel.__class__.__name__
            if typ == "ManyToOneRel" or typ == "OneToOneRel":
                kwargs['widget'] = w.VerboseForeignKeyRawIdWidget(db_field.rel, site, raw_id_fields_new_tab=self.raw_id_fields_new_tab)
            elif typ == "ManyToManyRel":
                kwargs['widget'] = w.VerboseManyToManyRawIdWidget(db_field.rel, site, raw_id_fields_new_tab=self.raw_id_fields_new_tab)
            return db_field.formfield(**kwargs)
        return super(BetterRawIdFieldsTabularInline, self).formfield_for_dbfield(db_field, **kwargs)

ImproveRawIdFieldsFormTabularInline = BetterRawIdFieldsTabularInline

class FormatterModelAdmin(BaseModelAdmin):
    """
    Allows the use of per-field formatters.

    Note, inheriting this class requires that the inheritor
    use rename their readonly_fields list to base_readonly_fields.
    This is necessary in order to dynamically insert formatters
    into the readonly_fields list to satisfy the model form field validation.
    """

    base_readonly_fields = ()

    @utils.classproperty
    def readonly_fields(cls):
        # Inserts our formatter instances into the readonly_field list.
        # We need to do this because admin/validation.py line ~243 uses
        # cls.readonly_fields instead of calling get_readonly_fields.
        readonly_fields = list(cls.base_readonly_fields)
        if cls.fieldsets:
            for title, data in cls.fieldsets: # pylint: disable=not-an-iterable
                for name in data['fields']:
                    if callable(name):
                        readonly_fields.append(name)
        elif cls.fields:
            for name in cls.fields: # pylint: disable=not-an-iterable
                if callable(name):
                    readonly_fields.append(name)
        return readonly_fields

    def get_readonly_fields(self, request, obj=None, check_fieldsets=True):
        # Inserts our formatter instances into the readonly_field list.
        readonly_fields = list(self.readonly_fields)
        #fieldsets = self.get_fieldsets(request, obj)
        fieldsets = self.declared_fieldsets
        if fieldsets:
            for title, data in fieldsets:
                for name in data['fields']:
                    if callable(name):
                        readonly_fields.append(name)
        return readonly_fields

class FormatterTabularInline(admin.TabularInline):

    base_readonly_fields = ()

    @utils.classproperty
    def readonly_fields(cls):
        # Inserts our formatter instances into the readonly_field list.
        # We need to do this because admin/validation.py line ~243 uses
        # cls.readonly_fields instead of calling get_readonly_fields.
        readonly_fields = list(cls.base_readonly_fields)
        for name in cls.fields: #pylint: disable=not-an-iterable
            if callable(name):
                readonly_fields.append(name)
        return readonly_fields+['id']

    def get_readonly_fields(self, request, obj=None):
        # Inserts our formatter instances into the readonly_field list.
        readonly_fields = list(self.readonly_fields)
        fieldsets = self.get_fieldsets(request, obj)
        for title, data in fieldsets:
            for name in data['fields']:
                if callable(name):
                    readonly_fields.append(name)
        return readonly_fields

    def id(self, request, obj=None):
        return obj.id

class ReadonlyModelAdmin(BaseModelAdmin):
    """
    Disables all delete or editing functionality in an admin view.
    """

    max_num = 0

    readonly_fields = ()

    def has_delete_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request, obj=None):
        return False

    def get_actions(self, request):
        actions = super(ReadonlyModelAdmin, self).get_actions(request)
        if 'delete_selected' in actions:
            del actions['delete_selected']
        return actions

    def get_readonly_fields(self, request, obj=None):
        readonly_fields = list(self.readonly_fields or [])
        exclude = list(self.exclude or [])
        for f in exclude:
            if f in readonly_fields:
                readonly_fields.remove(f)
        return readonly_fields + [f.name for f in self.model._meta.fields if f.name not in exclude]


class NoSaveModelForm(ModelForm):

    def save(self, force_insert=False, force_update=False, commit=True):
        return


class ReadonlyInlineModelAdminMixin(object):

    def get_readonly_fields(self, request, obj=None):
        lst = list(self.readonly_fields)
        lst.extend([f.name for f in self.model._meta.fields])
        return lst

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return True

    def has_delete_permission(self, request, obj=None):
        return False


class ReadonlyTabularInline(ReadonlyInlineModelAdminMixin, admin.TabularInline):
    form = NoSaveModelForm


class ReadonlyStackedInline(ReadonlyInlineModelAdminMixin, admin.StackedInline):
    form = NoSaveModelForm


def to_ascii(s):
    if not isinstance(s, six.string_types):
        return s
    return s.encode('ascii', errors='replace')


class CSVModelAdminMixin(object):
    """
    Adds a CSV export action to an admin view.
    """

    # This is the maximum number of records that will be written.
    # Be careful about increasing this.
    # Exporting massive numbers of records should be done asynchronously,
    # not in an admin request.
    csv_record_limit = 1000

    # If true, all fields from the queryset will be added to the results.
    csv_headers_all = False

    extra_csv_fields = ()

    # Specifies the header labels, of the form {data_field:header_label}.
    csv_header_names = {}

    csv_quoting = csv.QUOTE_MINIMAL

    # In cases where the default fields contain custom html meant for display in a webpage,
    # strip this out.
    csv_remove_html = True

    def get_actions(self, request):
        if hasattr(self, 'actions') and isinstance(self.actions, list):
            self.actions.append('csv_export')
        if isinstance(self, type) or (isclass(self) and issubclass(self, type)):
            return super(CSVModelAdmin, self).get_actions(request)

    def get_extra_csv_fields(self, request):
        return self.extra_csv_fields

    def get_csv_header_names(self, request):
        return self.csv_header_names

    def get_csv_raw_headers(self, request):
        if self.csv_headers_all:
            all_names = utils.get_model_fields(self.model)
            return all_names + list(self.get_extra_csv_fields(request))
        return list(self.list_display) + list(self.get_extra_csv_fields(request))

    def get_csv_queryset(self, request, qs):
        return qs

    def csv_export(self, request, qs=None, raw_headers=None):
        try:
            response = HttpResponse(mimetype='text/csv')
        except TypeError:
            response = HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = 'attachment; filename=%s.csv' \
            % slugify(self.model.__name__)

        if raw_headers is None:
            raw_headers = self.get_csv_raw_headers(request)

        def get_attr(obj, name, as_name=False):
            """
            Dereferences "__" delimited variable names.
            """
            parts = name.split('__')
            cursor = obj
            for part in parts:
                name = part
                cursor = getattr(cursor, part, None)
                if callable(cursor):
                    cursor = cursor()
            if cursor == obj:
                return
            if as_name:
                return name
            return cursor

        # Write header.
        header_data = {}
        fieldnames = []
        header_names = self.get_csv_header_names(request)

        # Write records.
        first = True
        qs = self.get_csv_queryset(request, qs)
        for r in qs[:self.csv_record_limit]:

            if first:
                first = False
                if not raw_headers:
                    if self.csv_headers_all and isinstance(r, dict):
                        if isinstance(qs, utils.DictCursor):
                            raw_headers = qs.field_order
                        else:
                            raw_headers = r.keys()
                    else:
                        raise Exception('No headers specified.')
                for name in raw_headers:
                    if name in header_names:
                        name_key = name
                        header_data[name] = header_names.get(name_key)
                    elif callable(name):
                        # This is likely a Formatter instance.
                        name_key = name.name
                        header_data[name_key] = name.short_description
                    elif isinstance(name, (tuple, list)) and len(name) == 2:
                        name_key, name_key_verbose = name
                        header_data[name_key] = name_key_verbose
                    elif isinstance(name, six.string_types) and hasattr(self, name):
                        # This is likely a ModelAdmin method name.
                        name_key = name
                        header_data[name_key] = getattr(self, name).short_description
                    elif hasattr(name, 'short_description'):
                        name_key = name
                        header_data[name_key] = getattr(
                            name, 'short_description')
                    elif hasattr(self.model, name):
                        name_key = name
                        if hasattr(getattr(self.model, name), 'short_description'):
                            header_data[name_key] = getattr(
                                getattr(self.model, name), 'short_description')
                        else:
                            header_data[name_key] = name
                    else:
                        name_key = name
                        header_data[name_key] = name_key#get_attr(r, name, as_name=True)
#                        field = self.model._meta.get_field_by_name(name)
#                        if field and field[0].verbose_name:
#                            header_data[name_key] = field[0].verbose_name
#                        else:
#                            header_data[name_key] = name
                    header_data[name_key] = header_data[name_key].title()
                    fieldnames.append(name_key)

                writer = csv.DictWriter(
                    response,
                    fieldnames=fieldnames,
                    quoting=self.csv_quoting)
                writer.writerow(header_data)
            #print('fieldnames:',fieldnames
            data = {}
            for name in raw_headers:
                obj = None
                if isinstance(r, dict):
                    if name in r:
                        data[name] = r[name]
    #                    print('skipping:',name
                        continue
#                    elif 'id' in r:
#                        obj = self.model.objects.get(id=r['id'])

                if callable(name):
                    # This is likely a Formatter instance.
                    name_key = name.name
                    if hasattr(name, 'plaintext'):
                        data[name_key] = to_ascii(name(r, plaintext=True))
                    else:
                        data[name_key] = to_ascii(name(r))
                elif isinstance(name, (tuple, list)) and len(name) == 2:
                    name_key, name_key_verbose = name
                    if hasattr(self, name_key):
                        data[name_key] = to_ascii(getattr(self, name_key))
                    else:
                        data[name_key] = to_ascii(getattr(r, name_key))
                elif isinstance(name, six.string_types) and hasattr(self, name):
                    # This is likely a ModelAdmin method name.
                    name_key = name
                    data[name_key] = to_ascii(getattr(self, name)(r))
                elif isinstance(name, six.string_types) and hasattr(r, name):
                    name_key = name
                    data[name_key] = to_ascii(getattr(r, name))
                else:
                    name_key = name
                    data[name_key] = to_ascii(get_attr(r, name))

                if callable(data[name_key]):
                    data[name_key] = to_ascii(data[name_key]())

                if self.csv_remove_html:
                    data[name_key] = utils.remove_html(data[name_key])

            #print('data:',data
            #writer.writerow(data)
            writer.writerow(utils.encode_csv_data(data))
        return response
    csv_export.short_description = \
        'Export selected %(verbose_name_plural)s as a CSV file'

class CSVModelAdmin(BaseModelAdmin, CSVModelAdminMixin):

    def get_actions(self, request):
        #TODO:is there a better way to do this? super() ignores the mixin's get_actions()...
        CSVModelAdminMixin.get_actions(self, request)
        return super(CSVModelAdmin, self).get_actions(request)

#https://djangosnippets.org/snippets/2484/
class LogEntryAdmin(ReadonlyModelAdmin):

    list_display = (
        'user',
        'action_time',
        'action',
        'admin_url',
    )

    list_filter = (
        filters.LogEntryAdminUserFilter,
    )

    def get_edited_object(self, obj):
        """
        Returns the edited object represented by this log entry.
        """
        model_class = obj.content_type.model_class()
        if model_class:
            obj = model_class.objects.get(id=obj.object_id)
            return obj

    def admin_url(self, obj=None):
        if not obj or not obj.id:
            return ''
        obj = self.get_edited_object(obj)
        if not obj:
            return ''
        if hasattr(obj, 'get_admin_url'):
            url = obj.get_admin_url()
        else:
            url = get_admin_change_url(obj)
        return mark_safe('<a href="%s">%s</a>' % (url, url))

    def action(self, obj):
        return str(obj)

    @classmethod
    def register(cls, admin_site=None):
        admin_site = admin_site or admin.site
        if hasattr(admin, 'models'):
            admin_site.register(admin.models.LogEntry, LogEntryAdmin)

#admin.site.register(models.LogEntry, LogEntryAdmin)




from django.db import models

class Person(models.Model):

    name = models.CharField(max_length=100, blank=False, null=False, unique=True)

class Contact(models.Model):

    person = models.ForeignKey(Person)

    email = models.EmailField(max_length=100, blank=False, null=False)

    class Meta:
        unique_together = (
            ('person', 'email'),
        )
